/*
 * #%L
 * JavaHg
 * %%
 * Copyright (C) 2011 aragost Trifork ag
 * %%
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * #L%
 */
package com.aragost.javahg.commands;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.aragost.javahg.internals.UnexpectedCommandOutputException;

/**
 * The statistic returned by running <tt>hg update</tt>.
 */
public class UpdateResult {

    private static final Pattern UPDATE_LINE_PATTERN = Pattern.compile("^(.*) files updated, (.*) files merged, (.*) files removed, (.*) files unresolved$");

    private int updated;

    private int merged;

    private int removed;

    private int unresolved;

    private UpdateResult() {
    }

    /**
     * @return number of updated files.
     */
    public int getUpdated() {
        return updated;
    }

    /**
     * @return number of merged files. These are the files that were
     *         merged without conflict, see {@link #getUnresolved()}
     *         for the number of files with merge conflicts.
     */
    public int getMerged() {
        return merged;
    }

    /**
     * @return number of removed files.
     */
    public int getRemoved() {
        return removed;
    }

    /**
     * @return number of unresolved files.
     */
    public int getUnresolved() {
        return unresolved;
    }

    /**
     * Factory method to create an {@link UpdateResult}.
     * 
     * @param line
     *            the final line of <tt>hg update</tt>
     * @return the parsed update result.
     */
    public static UpdateResult fromLine(String line) {
        // format of line: 0 files updated, 0 files merged, 0 files
        // removed, 0 files unresolved
        UpdateResult result = new UpdateResult();
        Matcher matcher = UPDATE_LINE_PATTERN.matcher(line);
        if (!matcher.matches()) {
            throw new UnexpectedCommandOutputException("Unexpected format: " + line);
        }
        int[] numbers = new int[4];
        for (int i = 0; i < 4; i++) {
            numbers[i] = Integer.parseInt(line.substring(matcher.start(i + 1), matcher.end(i + 1)));
        }
        result.updated = numbers[0];
        result.merged = numbers[1];
        result.removed = numbers[2];
        result.unresolved = numbers[3];
        return result;
    }

    @Override
    public String toString() {
        return "updated " + getUpdated() + ", merged " + getMerged() + ", removed " + getRemoved() + ", unresolved "
                + getUnresolved();
    }
}
